@@ -44,7 +44,14 @@ where
 
     #[emitter]
     fn emit_stylesheet(&mut self, n: &Stylesheet) -> Result {
-        self.emit_list(&n.rules, ListFormat::NotDelimited | ListFormat::MultiLine)?;
+        self.emit_list(
+            &n.rules,
+            if self.config.minify {
+                ListFormat::NotDelimited
+            } else {
+                ListFormat::NotDelimited | ListFormat::MultiLine
+            },
+        )?;
     }
 
     #[emitter]
@@ -59,9 +66,7 @@ where
     #[emitter]
     fn emit_qualified_rule(&mut self, n: &QualifiedRule) -> Result {
         emit!(self, n.prelude);
-        if !self.config.minify {
-            space!(self);
-        }
+        formatting_space!(self);
         emit!(self, n.block);
     }
 
@@ -79,6 +84,7 @@ where
             AtRule::Namespace(n) => emit!(self, n),
             AtRule::Viewport(n) => emit!(self, n),
             AtRule::Document(n) => emit!(self, n),
+            AtRule::ColorProfile(n) => emit!(self, n),
             AtRule::Unknown(n) => emit!(self, n),
         }
     }
@@ -96,6 +102,10 @@ where
         punct!(self, "@");
         keyword!(self, "charset");
 
+        // https://drafts.csswg.org/css2/#charset%E2%91%A0
+        // @charset must be written literally, i.e., the 10 characters '@charset "'
+        // (lowercase, no backslash escapes), followed by the encoding name, followed by
+        // ";.
         space!(self);
 
         emit!(self, n.charset);
@@ -107,16 +117,31 @@ where
     fn emit_import_rule(&mut self, n: &ImportRule) -> Result {
         punct!(self, "@");
         keyword!(self, "import");
-        space!(self);
+
+        match n.href {
+            ImportHref::Url(_) => {
+                space!(self);
+            }
+            ImportHref::Str(_) => {
+                formatting_space!(self);
+            }
+        }
+
         emit!(self, n.href);
 
         if let Some(layer_name) = &n.layer_name {
-            space!(self);
+            formatting_space!(self);
             emit!(self, layer_name);
+
+            if self.config.minify && (n.supports.is_some() || n.media.is_some()) {
+                if let ImportLayerName::Ident(_) = layer_name {
+                    space!(self);
+                }
+            }
         }
 
         if let Some(supports) = &n.supports {
-            space!(self);
+            formatting_space!(self);
             keyword!(self, "supports");
             punct!(self, "(");
             emit!(self, supports);
@@ -124,7 +149,7 @@ where
         }
 
         if let Some(media) = &n.media {
-            space!(self);
+            formatting_space!(self);
             emit!(self, media);
         }
 
@@ -151,8 +176,7 @@ where
     fn emit_font_face_rule(&mut self, n: &FontFaceRule) -> Result {
         punct!(self, "@");
         keyword!(self, "font-face");
-        space!(self);
-
+        formatting_space!(self);
         emit!(self, n.block);
     }
 
@@ -168,13 +192,32 @@ where
     fn emit_keyframes_rule(&mut self, n: &KeyframesRule) -> Result {
         punct!(self, "@");
         keyword!(self, "keyframes");
-        space!(self);
+
+        match n.name {
+            KeyframesName::Str(_) => {
+                formatting_space!(self);
+            }
+            KeyframesName::CustomIdent(_) => {
+                space!(self);
+            }
+        }
 
         emit!(self, n.name);
-        space!(self);
 
+        formatting_space!(self);
         punct!(self, "{");
-        self.emit_list(&n.blocks, ListFormat::NotDelimited)?;
+        formatting_newline!(self);
+
+        self.emit_list(
+            &n.blocks,
+            if self.config.minify {
+                ListFormat::NotDelimited
+            } else {
+                ListFormat::MultiLine
+            },
+        )?;
+
+        formatting_newline!(self);
         punct!(self, "}");
     }
 
@@ -200,15 +243,24 @@ where
     fn emit_layer_rule(&mut self, n: &LayerRule) -> Result {
         punct!(self, "@");
         keyword!(self, "layer");
-        space!(self);
 
         if n.prelude.is_some() {
+            space!(self);
             emit!(self, n.prelude);
+        } else {
+            formatting_space!(self);
         }
 
         if let Some(rules) = &n.rules {
             punct!(self, "{");
-            self.emit_list(rules, ListFormat::NotDelimited | ListFormat::MultiLine)?;
+            self.emit_list(
+                rules,
+                if self.config.minify {
+                    ListFormat::NotDelimited
+                } else {
+                    ListFormat::NotDelimited | ListFormat::MultiLine
+                },
+            )?;
             punct!(self, "}");
         } else {
             punct!(self, ";");
@@ -219,7 +271,7 @@ where
     fn emit_keyframe_block(&mut self, n: &KeyframeBlock) -> Result {
         self.emit_list(&n.selector, ListFormat::CommaDelimited)?;
 
-        space!(self);
+        formatting_space!(self);
 
         emit!(self, n.rule);
     }
@@ -240,7 +292,7 @@ where
 
         emit!(self, n.media);
 
-        space!(self);
+        formatting_space!(self);
 
         punct!(self, "{");
         self.emit_list(&n.rules, ListFormat::NotDelimited | ListFormat::MultiLine)?;
@@ -382,7 +434,7 @@ where
     fn emit_media_feature_plain(&mut self, n: &MediaFeaturePlain) -> Result {
         emit!(self, n.name);
         punct!(self, ":");
-        space!(self);
+        formatting_space!(self);
         emit!(self, n.value);
     }
 
@@ -421,7 +473,7 @@ where
 
         emit!(self, n.condition);
 
-        space!(self);
+        formatting_space!(self);
 
         punct!(self, "{");
         self.emit_list(&n.rules, ListFormat::NotDelimited)?;
@@ -532,7 +584,7 @@ where
     fn emit_viewport_rule(&mut self, n: &ViewportRule) -> Result {
         punct!(self, "@");
         keyword!(self, "viewport");
-        space!(self);
+        formatting_space!(self);
 
         emit!(self, n.block);
     }
@@ -545,19 +597,49 @@ where
 
         self.emit_list(&n.selectors, ListFormat::CommaDelimited)?;
 
-        space!(self);
+        formatting_space!(self);
 
         punct!(self, "{");
         self.emit_list(&n.block, ListFormat::NotDelimited)?;
         punct!(self, "}");
     }
 
+    fn emit_list_values(&mut self, nodes: &[Value], format: ListFormat) -> Result {
+        let mut need_space = true;
+
+        for (idx, node) in nodes.iter().enumerate() {
+            if idx != 0 {
+                match node {
+                    Value::Delimiter(Delimiter {
+                        value: DelimiterValue::Comma,
+                        ..
+                    }) => {
+                        need_space = !self.config.minify;
+                    }
+                    _ => {
+                        if need_space {
+                            self.write_delim(format)?;
+                        } else {
+                            need_space = true;
+                        }
+                    }
+                };
+            }
+
+            emit!(self, node)
+        }
+
+        Ok(())
+    }
+
     #[emitter]
     fn emit_function(&mut self, n: &Function) -> Result {
         emit!(self, n.name);
-
         punct!(self, "(");
-        self.emit_list(&n.value, ListFormat::CommaDelimited)?;
+        self.emit_list_values(
+            &n.value,
+            ListFormat::SpaceDelimited | ListFormat::SingleLine,
+        )?;
         punct!(self, ")");
     }
 
@@ -572,16 +654,43 @@ where
             Value::Ratio(n) => emit!(self, n),
             Value::Hash(n) => emit!(self, n),
             Value::Ident(n) => emit!(self, n),
+            Value::DashedIdent(n) => emit!(self, n),
             Value::Str(n) => emit!(self, n),
             Value::Bin(n) => emit!(self, n),
-            Value::Space(n) => emit!(self, n),
             Value::Tokens(n) => emit!(self, n),
             Value::AtText(n) => emit!(self, n),
             Value::Url(n) => emit!(self, n),
-            Value::Comma(n) => emit!(self, n),
+            Value::Delimiter(n) => emit!(self, n),
         }
     }
 
+    #[emitter]
+    fn emit_at_rule_name(&mut self, n: &AtRuleName) -> Result {
+        match n {
+            AtRuleName::Ident(n) => emit!(self, n),
+            AtRuleName::DashedIdent(n) => emit!(self, n),
+        }
+    }
+
+    #[emitter]
+    fn emit_color_profile_name(&mut self, n: &ColorProfileName) -> Result {
+        match n {
+            ColorProfileName::Ident(n) => emit!(self, n),
+            ColorProfileName::DashedIdent(n) => emit!(self, n),
+        }
+    }
+
+    #[emitter]
+    fn emit_color_profile_rule(&mut self, n: &ColorProfileRule) -> Result {
+        punct!(self, "@");
+        keyword!(self, "color-profile");
+        space!(self);
+        emit!(self, n.name);
+        punct!(self, "{");
+        self.emit_list(&n.block, ListFormat::NotDelimited)?;
+        punct!(self, "}");
+    }
+
     #[emitter]
     fn emit_unknown_at_rule(&mut self, n: &UnknownAtRule) -> Result {
         punct!(self, "@");
@@ -628,7 +737,14 @@ where
     fn emit_block(&mut self, n: &Block) -> Result {
         punct!(self, "{");
 
-        self.emit_list(&n.value, ListFormat::SemiDelimited | ListFormat::MultiLine)?;
+        self.emit_list(
+            &n.value,
+            if self.config.minify {
+                ListFormat::SemiDelimited
+            } else {
+                ListFormat::SemiDelimited | ListFormat::MultiLine
+            },
+        )?;
 
         punct!(self, "}");
     }
@@ -642,23 +758,36 @@ where
         }
     }
 
+    #[emitter]
+    fn emit_declaration_property(&mut self, n: &DeclarationProperty) -> Result {
+        match n {
+            DeclarationProperty::Ident(n) => emit!(self, n),
+            DeclarationProperty::DashedIdent(n) => emit!(self, n),
+        }
+    }
+
     #[emitter]
     fn emit_declaration(&mut self, n: &Declaration) -> Result {
         emit!(self, n.property);
         punct!(self, ":");
 
-        let is_custom_property = n.property.value.starts_with("--");
+        let is_custom_property = match n.property {
+            DeclarationProperty::DashedIdent(_) => true,
+            DeclarationProperty::Ident(_) => false,
+        };
 
         if !is_custom_property {
             formatting_space!(self);
         }
 
-        let format = match is_custom_property {
-            true => ListFormat::NotDelimited,
-            false => ListFormat::SpaceDelimited | ListFormat::SingleLine,
-        };
-
-        self.emit_list(&n.value, format)?;
+        if is_custom_property {
+            self.emit_list(&n.value, ListFormat::NotDelimited)?;
+        } else {
+            self.emit_list_values(
+                &n.value,
+                ListFormat::SpaceDelimited | ListFormat::SingleLine,
+            )?;
+        }
 
         if let Some(tok) = n.important {
             if !is_custom_property {
@@ -685,6 +814,11 @@ where
         self.wr.write_raw(Some(n.span), &n.raw)?;
     }
 
+    #[emitter]
+    fn emit_dashed_ident(&mut self, n: &DashedIdent) -> Result {
+        self.wr.write_raw(Some(n.span), &n.raw)?;
+    }
+
     #[emitter]
     fn emit_keyframe_block_rule(&mut self, n: &KeyframeBlockRule) -> Result {
         match n {
@@ -702,8 +836,7 @@ where
     #[emitter]
     fn emit_page_rule_block(&mut self, n: &PageRuleBlock) -> Result {
         punct!(self, "{");
-
-        self.wr.write_newline()?;
+        formatting_newline!(self);
 
         self.wr.increase_indent();
 
@@ -713,8 +846,7 @@ where
         };
         self.with_ctx(ctx)
             .emit_list(&n.items, ListFormat::MultiLine | ListFormat::NotDelimited)?;
-
-        self.wr.write_newline()?;
+        formatting_newline!(self);
 
         self.wr.decrease_indent();
 
@@ -767,13 +899,8 @@ where
     }
 
     #[emitter]
-    fn emit_comma_values(&mut self, n: &CommaValues) -> Result {
-        self.emit_list(&n.values, ListFormat::CommaDelimited)?;
-    }
-
-    #[emitter]
-    fn emit_space_values(&mut self, n: &SpaceValues) -> Result {
-        self.emit_list(&n.values, ListFormat::SpaceDelimited)?;
+    fn emit_delimiter(&mut self, n: &Delimiter) -> Result {
+        punct!(self, n.value.as_str());
     }
 
     #[emitter]
@@ -899,7 +1026,10 @@ where
         }
 
         if let Some(modifiers) = &n.modifiers {
-            self.emit_list(modifiers, ListFormat::SpaceDelimited)?;
+            if !modifiers.is_empty() {
+                formatting_space!(self);
+                self.emit_list(modifiers, ListFormat::SpaceDelimited)?;
+            }
         }
 
         punct!(self, ")");
@@ -1052,7 +1182,15 @@ where
         emit!(self, n.value);
 
         if let Some(m) = &n.modifier {
-            space!(self);
+            match n.value {
+                Some(AttrSelectorValue::Str(_)) => {
+                    formatting_space!(self);
+                }
+                Some(AttrSelectorValue::Ident(_)) => {
+                    space!(self);
+                }
+                _ => {}
+            }
 
             self.wr.write_raw_char(None, *m)?;
         }
@@ -1140,9 +1278,10 @@ where
                 self.write_delim(format)?;
 
                 if format & ListFormat::LinesMask == ListFormat::MultiLine {
-                    self.wr.write_newline()?;
+                    formatting_newline!(self);
                 }
             }
+
             emit!(self, node)
         }
 
@@ -1154,7 +1293,7 @@ where
             ListFormat::None => {}
             ListFormat::CommaDelimited => {
                 punct!(self, ",");
-                space!(self);
+                formatting_space!(self);
             }
             ListFormat::SpaceDelimited => {
                 space!(self)
